@@ -1,5 +1,5 @@
 /*
- * Copyright © 2018, VideoLAN and dav1d authors
+ * Copyright © 2018-2021, VideoLAN and dav1d authors
  * Copyright © 2018, Two Orioles, LLC
  * All rights reserved.
  *
@@ -151,6 +151,8 @@ struct Dav1dContext {
     int all_layers;
     unsigned frame_size_limit;
     int drain;
+    enum PictureFlags frame_flags;
+    enum Dav1dEventFlags event_flags;
 
     Dav1dLogger logger;
 
